<!doctype html>
{% load static %}
{% load crispy_forms_tags %}

<html lang="en">
<head>
    <!-- Required meta tags -->
    <meta charset="utf-8">
    <meta name="viewport" content="width=device-width, initial-scale=1, shrink-to-fit=no">

    {% block extra_head %}
        <!-- Bootstrap CSS -->
        <link href="https://cdn.jsdelivr.net/npm/bootstrap@5.0.2/dist/css/bootstrap.min.css" rel="stylesheet"
              integrity="sha384-EVSTQN3/azprG1Anm3QDgpJLIm9Nao0Yz1ztcQTwFspd3yD65VohhpuuCOmLASjC"
              crossorigin="anonymous">
        <script src="https://cdn.jsdelivr.net/npm/bootstrap@5.0.2/dist/js/bootstrap.bundle.min.js"
                integrity="sha384-MrcW6ZMFYlzcLA8Nl+NtUVF0sA7MsXsP1UyJoMp4YLEuNSfAP+JcXn/tWtIaxVXM"
                crossorigin="anonymous"></script>
    {% endblock %}

    <title>{{ report_title }} | Django Slick Reporting</title>
</head>
<body>
<div class="container">


    <div class="py-5 ">
        <h2 class="page-title"> {{ report_title }} </h2>
        {% block content %}

        {% endblock %}
    </div>
</div>



{% block extrajs %}
    {% include "slick_reporting/js_resources.html" %}

{% endblock %}
</body>
</html>